from unittest import mock
from redis import Redis

from assemblyline.odm.messages.task import Task

from assemblyline.odm.random_data import random_model_obj
from assemblyline.odm.models.service import Service

from assemblyline_core.plumber.run_plumber import Plumber
from assemblyline_core.server_base import ServiceStage

from mocking import TrueCountTimes


def test_expire_missing_service():
    redis = mock.MagicMock(spec=Redis)
    redis.keys.return_value = [b'service-queue-not-service-a']
    redis.zcard.return_value = 0
    redis_persist = mock.MagicMock(spec=Redis)
    datastore = mock.MagicMock()

    service_a = random_model_obj(Service)
    service_a.name = 'a'
    service_a.enabled = True

    datastore.list_all_services.return_value = [service_a]
    plumber = Plumber(redis=redis, redis_persist=redis_persist, datastore=datastore, delay=1)
    plumber.get_service_stage = mock.MagicMock(return_value=ServiceStage.Running)
    plumber.dispatch_client = mock.MagicMock()

    task = random_model_obj(Task)
    plumber.dispatch_client.request_work.side_effect = [task, None, None]

    plumber.running = TrueCountTimes(count=1)
    plumber.service_queue_plumbing()

    assert plumber.dispatch_client.service_failed.call_count == 1
    args = plumber.dispatch_client.service_failed.call_args
    assert args[0][0] == task.sid


def test_flush_paused_queues():
    redis = mock.MagicMock(spec=Redis)
    redis.keys.return_value = [b'service-queue-a']
    redis.zcard.return_value = 0
    redis_persist = mock.MagicMock(spec=Redis)
    datastore = mock.MagicMock()

    service_a = random_model_obj(Service)
    service_a.name = 'a'
    service_a.enabled = True

    datastore.list_all_services.return_value = [service_a]

    plumber = Plumber(redis=redis, redis_persist=redis_persist, datastore=datastore, delay=1)
    plumber.get_service_stage = mock.MagicMock(return_value=ServiceStage.Running)
    plumber.dispatch_client = mock.MagicMock()

    task = random_model_obj(Task)
    plumber.dispatch_client.request_work.side_effect = [task, None, None]

    plumber.running = TrueCountTimes(count=1)
    plumber.service_queue_plumbing()

    assert plumber.dispatch_client.service_failed.call_count == 0

    plumber.get_service_stage = mock.MagicMock(return_value=ServiceStage.Paused)

    plumber.running = TrueCountTimes(count=1)
    plumber.service_queue_plumbing()

    assert plumber.dispatch_client.service_failed.call_count == 1
    args = plumber.dispatch_client.service_failed.call_args
    assert args[0][0] == task.sid



import shutil
import os

from assemblyline_core.replay.client import APIClient, DirectClient
from assemblyline_core.replay.replay import ReplayBase


class ReplayLoaderWorker(ReplayBase):
    def __init__(self):
        super().__init__("assemblyline.replay_loader.worker")

        # Load client
        if self.replay_config.loader.client.type == 'direct':
            self.log.info("Using direct database access client")
            self.client = DirectClient(self.log)
        elif self.replay_config.loader.client.type == 'api':
            self.log.info(f"Using API access client to ({self.replay_config.loader.client.options.host})")
            self.client = APIClient(self.log, **self.replay_config.loader.client.options.as_primitives())
        else:
            raise ValueError(f'Invalid client type ({self.replay_config.loader.client.type}). '
                             'Must be either \'api\' or \'direct\'.')

    def process_file(self, once=False):
        while self.running:
            file_path = self.client.get_next_file()

            if file_path:
                self.log.info(f"Processing file: {file_path}")
                try:
                    self.client.load_bundle(file_path,
                                            min_classification=self.replay_config.loader.min_classification,
                                            rescan_services=self.replay_config.loader.rescan)
                    if os.path.exists(file_path):
                        os.unlink(file_path)
                except OSError as e:
                    # Critical exception occurred
                    if 'Stale file handle' in str(e):
                        # Terminate on stale file handle from NFS mount
                        self.log.warning("Stale file handle detected. Terminating..")
                        self.stop()
                    elif 'Invalid cross-device link' in str(e):
                        # Terminate on NFS-related error
                        self.log.warning("'Invalid cross-device link' exception detected. Terminating..")
                        self.stop()
                except Exception:
                    # Make sure failed directory exists
                    os.makedirs(self.replay_config.loader.failed_directory, exist_ok=True)

                    self.log.error(f"Failed to load the bundle file {file_path}, moving it to the failed directory.")
                    failed_path = os.path.join(self.replay_config.loader.failed_directory, os.path.basename(file_path))
                    shutil.move(file_path, failed_path)

            if once:
                break

    def try_run(self):
        threads = {}

        for ii in range(self.replay_config.loader.input_threads):
            threads[f'File processor #{ii}'] = self.process_file

        self.maintain_threads(threads)

    def stop(self):
        return super().stop()


if __name__ == '__main__':
    with ReplayLoaderWorker() as replay:
        replay.serve_forever()
